/*
    SAW_pivot implements the pivot algorithm for Monte Carlo simulations
      of self-avoiding walks
    Copyright (C) 2003 Tom Kennedy

    This file is part of SAW_pivot (version 1.0).

    SAW_pivot is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SAW_pivot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SAW_pivot; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    E-mail contact: tgk@math.arizona.edu
    Ordinary mail: 
        Tom Kennedy
        Mathematics Department
        University of Arizona
        Tucson, ZA 85721, USA

*/

// file name : src/point.c
//
// many functions for the class point are defined in point.h since they 
// are inlined.
//

#include <cstdio> 
using namespace std; 
#include "math.h"
#include <iostream>
using namespace std; 
#include <cstdlib>
#include "sys/time.h"
#include "time.h"
#include "limits.h"
#include <complex>
using namespace std;

#define I complex<double>(0.,1.);

#include "local_constants.h"
#include "rpoint.h"
#include "point.h"


extern long group_product[NUM_SYM+1][NUM_SYM+1];
extern long group_inverse[NUM_SYM+1];

////////////////////////////////////////////////////////////////
//             member functions for class point               //
////////////////////////////////////////////////////////////////

long point::coord_x_slow() {return(x);}
long point::coord_y_slow() {return(y);}
#ifdef THREE_DIMENSIONS
long point::coord_z_slow() {return(z);}
#endif
#ifdef FOUR_DIMENSIONS
long point::coord_z_slow() {return(z);}
long point::coord_w_slow() {return(w);}
#endif

#ifdef TWO_DIMENSIONS
void point::assign(long xx,long yy) {x=xx; y=yy;} 
#endif
#ifdef THREE_DIMENSIONS
void point::assign(long xx,long yy,long zz) {x=xx; y=yy; z=zz;} 
#endif
#ifdef FOUR_DIMENSIONS
void point::assign(long xx,long yy,long zz,long ww) {x=xx; y=yy; z=zz; w=ww;} 
#endif

void point::zero() 
{
    x=0; y=0;
    #ifdef THREE_DIMENSIONS
        z=0;
    #endif
    #ifdef FOUR_DIMENSIONS
        z=0;
        w=0;
    #endif
} 

double point::distance()  
// returns euclidean distance from point to origin. Used by walk::turn_frac() 
{
    rpoint rp;
    rp=(*this);
    return(rp.distance());
} // end point::distance()

void point::print(FILE *fptr) 
{
    #ifdef TWO_DIMENSIONS
        fprintf(fptr,"%ld %ld \n",x,y);
    #endif
    #ifdef THREE_DIMENSIONS
        fprintf(fptr,"%ld %ld %ld\n",x,y,z);
    #endif
    #ifdef FOUR_DIMENSIONS
        fprintf(fptr,"%ld %ld %ld %ld\n",x,y,z,w);
    #endif
}

double point::real_coord_x() 
// returns real x coordinate of point, rather than coords wrt lattice structure
{
    #ifdef SQUARE_LATTICE
        return(x);
    #endif

    #ifdef HEXAGONAL_TRIANGULAR_LATTICE
        return(x-0.5*y);
    #endif

    #ifdef CUBIC_LATTICE
        return(x);
    #endif

    #ifdef HYPER_LATTICE
        return(x);
    #endif

    #ifdef PERCOLATION_HEXAGONAL_LATTICE
        return((x-y)*sqrt(3.)/2.);
    #endif

    #ifdef ISING_TRIANGULAR_LATTICE
        return((x-y)*sqrt(3.)/2.);
    #endif

    #ifdef ISING_SQUARE_LATTICE
        return(x);
    #endif

    printf("error in real_coord_x - bad case \n");
    exit(1);
} // end real_coord_x()

double point::real_coord_y()
// returns real y coordinate of point, rather than coords wrt lattice structure
{
    #ifdef SQUARE_LATTICE
        return(y);
    #endif

    #ifdef HEXAGONAL_TRIANGULAR_LATTICE
        return(0.5*sqrt(3.)*y);
    #endif

    #ifdef CUBIC_LATTICE
        return(y);
    #endif

    #ifdef HYPER_LATTICE
        return(y);
    #endif

    #ifdef PERCOLATION_HEXAGONAL_LATTICE
        return((x+y)*0.5);
    #endif

    #ifdef ISING_TRIANGULAR_LATTICE
        return((x+y)*0.5);
    #endif

    #ifdef ISING_SQUARE_LATTICE
        return(y);
    #endif

    printf("error in real_coord_y - bad case \n");
    exit(1);
} // end real_coord_y()

#ifdef THREE_DIMENSIONS
double point::real_coord_z()  
// returns real z coordinate of point, rather than coords wrt lattice structure
{
    #ifdef CUBIC_LATTICE
        return(z);
    #endif

    #ifdef HYPER_LATTICE
        return(z);
    #endif

    printf("error in real_coord_z - bad case \n");
    exit(1);

} // end real_coord_z()
#endif

#ifdef FOUR_DIMENSIONS
double point::real_coord_z() 
// returns real z coordinate of point, rather than coords wrt lattice structure
{
    #ifdef HYPER_LATTICE
        return(z);
    #endif

    printf("error in real_coord_z - bad case \n");
    exit(1);

} // end real_coord_z()

double point::real_coord_w() 
// returns real w coordinate of point, rather than coords wrt lattice structure
{
    #ifdef HYPER_LATTICE
        return(w);
    #endif

    printf("error in real_coord_w - bad case \n");
    exit(1);

} // end real_coord_w()
#endif

point& point::operator*=(long c) 
{
    x*=c;
    y*=c;
    #ifdef THREE_DIMENSIONS
        z*=c;
    #endif
    #ifdef FOUR_DIMENSIONS
        z*=c;
        w*=c;
    #endif
    return *this;
} 

point operator*(long c,point p) 
{
    point r=p; 
    return r*=c;
} 

void point::scan(FILE *fptr) 
{
    #ifdef TWO_DIMENSIONS
        fscanf(fptr,"%ld %ld",&x,&y);
    #endif
    #ifdef THREE_DIMENSIONS
        fscanf(fptr,"%ld %ld %ld",&x,&y,&z);
    #endif
    #ifdef FOUR_DIMENSIONS
        fscanf(fptr,"%ld %ld %ld %ld",&x,&y,&z,&w);
    #endif
}

int operator==(point p,point q)
{
    #ifdef TWO_DIMENSIONS
        if (p.coord_x()!=q.coord_x() || p.coord_y()!=q.coord_y()) return(0);
    #endif
    #ifdef THREE_DIMENSIONS
        if (p.coord_x()!=q.coord_x() || p.coord_y()!=q.coord_y()
            || p.coord_z()!=q.coord_z()) return(0);
    #endif
    #ifdef FOUR_DIMENSIONS
        if (p.coord_x()!=q.coord_x() || p.coord_y()!=q.coord_y()
            || p.coord_z()!=q.coord_z() || p.coord_w()!=q.coord_w()) return(0);
    #endif
    return(1);
} 

point& point::operator=(const point& p) 
{
    x=p.x;
    y=p.y;
    #ifdef THREE_DIMENSIONS
        z=p.z;
    #endif
    #ifdef FOUR_DIMENSIONS
        z=p.z;
        w=p.w;
    #endif
    return(*this);
} // end point::operator=()


long operator*(point p,point q)
// return p x q
{
    #ifdef TWO_DIMENSIONS
        return(p.x*q.y-p.y*q.x);
    #endif
    return(0);    
}


///////////////////////////////////////////////////////////////
//  checkk checkk checkk checkk checkk checkk checkk checkk  //
///////////////////////////////////////////////////////////////
