// file name : src/point.h
//

#ifndef __point_hh
#define __point_hh

// Points are represented by two classes : point and rpoint.
// rpoint contains real coordinates and represents the actual coordinates of
// the point in R^d.
// point has integer coordinates. These are the coordinates of the point
// with respect to a basis. The bases used are 
// square :        e1=(1,0)  e2=(0,1)
// triangular :    e1=(1,0)  e2=(-1/2,root(3)/2)
// hexagonal :     e1=(1,0)  e2=(-1/2,root(3)/2)
// cubic :         e1=(1,0,0) e2=(0,1,0)  e3=(0,0,1)
// hypercubic :    e1=(1,0,0,0) e2=(0,1,0,0) e3=(0,0,1,0) e4=(0,0,0,1)  
// percolate hex:  e1=(root(3)/2,1/2) e1=(-root(3)/2,1/2)
// ising tri:      e1=(root(3)/2,1/2) e1=(-root(3)/2,1/2)
// ising square:   e1=(1,0)  e2=(0,1)
class excluded_class;

class point {
private:
public:
    // these are coordinates wrt the two vectors that define the lattice
    #ifdef TWO_DIMENSIONS
        long x,y;
    #endif
    #ifdef THREE_DIMENSIONS
        long x,y,z;
    #endif
    #ifdef FOUR_DIMENSIONS
        long x,y,z,w;
    #endif
    inline point& operator+=(point p) 
    {
        x+=p.x;
        y+=p.y;
        #ifdef THREE_DIMENSIONS
            z+=p.z;
        #endif
        #ifdef FOUR_DIMENSIONS
            z+=p.z;
            w+=p.w;
        #endif
        return *this;
    } 

    inline point& operator-=(point p) 
    {
        x-=p.x;
        y-=p.y;
        #ifdef THREE_DIMENSIONS
            z-=p.z;
        #endif
        #ifdef FOUR_DIMENSIONS
           z-=p.z;
           w-=p.w;
        #endif
        return *this;
    } 

    point& operator*=(long c);
    void print(FILE *fptr);
    void scan(FILE *fptr);
    double distance(); // distance to the origin
    point& operator=(const point& p); 
    void zero(); // sets coords to zero 

    inline long lattice_distance() 
    // computes minimum number of steps to get from point with integer
    // coefs (x,y) to origin, or a lower bound on this number.
    // NB : For hexagonal lattice we use the separation on the triangular 
    //   lattice. This is valid since the triangular latttice separation will
    //   be a lower bound on the hexagonal lattice separation, but it 
    //   may not be the fastest.
    {
        long separation=0;
        separation+=0; // to avoid compiler warning
 
        // in this routine we use if then statement in place of abs - faster

        #ifdef PERCOLATION_HEXAGONAL_LATTICE // only to avoid compiler warning
            x=0;
            y=0;
        #endif

        #ifdef SQUARE_LATTICE
            if (x<0) separation-=x;
            else separation+=x;
            if (y<0) separation-=y;
            else separation+=y;
            return(separation);
        #endif


        // hexagonal just uses triangular lattice separation
        // percolation hexagonal just uses triangular lattice separation
        #if defined (HEXAGONAL_TRIANGULAR_LATTICE) || defined (PERCOLATION_HEXAGONAL_LATTICE)
            // Point can be written in three ways:
            // x e1 + y e2
            // x (e2+e1) + (y-x) e2
            // (x-y) e1 + y (e1+e2)
            // dif1,dif2, dif3 are |  | of x,y,x-y. The min distance is min of 
            // dif1+dif2, dif1+dif3, dif2+dif3.

            long temp,dif1,dif2,dif3;
            if (x>0) dif1=x;
            else dif1=-x;
            if (y>0) dif2=y;
            else dif2=-y;
            dif3=x-y;
            if (dif3<0) dif3=-dif3;

            separation=dif1+dif2;
            temp=dif3+dif1;
            if (temp<separation) separation=temp;
            temp=dif3+dif2;
            if (temp<separation) separation=temp;
            return(separation);
        #endif

        #ifdef CUBIC_LATTICE
            if (x<0) separation-=x;
            else separation+=x;
            if (y<0) separation-=y;
            else separation+=y;
            if (z<0) separation-=z;
            else separation+=z;
            return(separation);
        #endif
  
        #ifdef HYPER_LATTICE
            if (x<0) separation-=x;
            else separation+=x;
            if (y<0) separation-=y;
            else separation+=y;
            if (z<0) separation-=z;
            else separation+=z;
            if (w<0) separation-=w;
            else separation+=w;
            return(separation);
        #endif

        printf("error in lattice_distance - bad case \n");
        exit(1);
 
    } // end lattice_distance()

    inline void euclidean_op(point* q,point* p,long isym) 
    // isym specifies the group element g. Routine compute gq+p
    {
  
        #ifdef SQUARE_LATTICE
            // The order of the group elements is important here. 
            // algorithm=1 and algorithm=2 take advantage of the fact that 
            // reflections in 45 deg lines are 1 and 2 in the list. 
            switch(isym) {
                case 0: x=(*q).x;  y=(*q).y;  break; // identity
                case 1: x=(*q).y;  y=(*q).x;  break; // reflect in 45 deg line
                case 2: x=-(*q).y; y=-(*q).x; break; // reflect in -45 deg line
                case 3: x=(*q).y;  y=-(*q).x; break; // rotate by 90 clockwise
                case 4: x=-(*q).x; y=-(*q).y; break; // rotate by 180 clockwise
                case 5: x=-(*q).y; y=(*q).x;  break; // rotate by 270 clockwise
                case 6: x= (*q).x; y=-(*q).y; break; // reflect in x-axis 
                case 7: x=-(*q).x; y=(*q).y;  break; // reflect in y-axis 
                default:
                    printf("bad case square:euclidean_op, isym=%ld \n",isym);
                    exit(0);
                break;
	} // end switch on isym
        #endif

        #ifdef HEXAGONAL_TRIANGULAR_LATTICE
            // rotations here are counterclockwise
            switch(isym) {
                 // identity
                case 0:   x=(*q).x;      y=(*q).y;          break;  
                // R^2, rotate by 120
                case 1:   x=-(*q).y;     y= (*q).x-(*q).y; break;   
                // R^4, rotate by 240
                case 2:   x=-(*q).x+(*q).y; y=-(*q).x;     break;   
                // F_1, flip about e1 
                case 3:   x= (*q).x-(*q).y; y=-(*q).y;     break;  
                // F_2, flip about e2
                case 4:   x=-(*q).x;     y=-(*q).x+(*q).y; break;   
                // F_3, flip about e1+e2
                case 5:   x= (*q).y;     y= (*q).x;     break;   
                // R, rotation by 60
                case 6:   x= (*q).x-(*q).y; y= (*q).x;     break;   
                // R^3, rotation by 180
                case 7:   x=-(*q).x;         y=-(*q).y;     break;    
                // R^5, rotation by 300
                case 8:   x= (*q).y;     y=-(*q).x+(*q).y; break;  
                // -F_1
                case 9:   x=-(*q).x+(*q).y; y= (*q).y;     break;  
                // -F_2
                case 10:  x= (*q).x;     y= (*q).x-(*q).y; break;   
                // -F_3
                case 11:  x=-(*q).y;     y=-(*q).x;     break;   
                default:
                    printf("bad case hex_tri in euclidean_op \n");
                    exit(0);
                break;
            } // end switch on isym
        #endif
 
        #if ALGORITHM_NUMBER==47
            switch(isym%6) {
                case 0:   x=(*q).x;   y=(*q).y;   z=(*q).z;  break; // identity
                case 1:   x=(*q).y;   y=(*q).x;   z=(*q).z;  break;   // (xy)
                case 2:   x=(*q).x;   y=(*q).z;   z=(*q).y;  break;   // (yz)
                case 3:   x=(*q).z;   y=(*q).y;   z=(*q).x;  break;   // (xz)
                case 4:   x=(*q).z;   y=(*q).x;   z=(*q).y;  break;   // (xyz)
                case 5:   x=(*q).y;   y=(*q).z;   z=(*q).x;  break;   // (xzy) 
                default:  
                    printf("bad case cubic isym_perm in euclidean_op \n");
                    exit(0);
                break;
            } // end switch on isym_perm
            switch(isym/6) {
                case 0:                        break;   // +++
                case 1:                 z=-z;  break;   // ++-
                case 2:          y=-y;         break;   // +-+
                case 3:          y=-y;  z=-z;  break;   // +--
                case 4:   x=-x;                break;   // -++
                case 5:   x=-x;         z=-z;  break;   // -+-
                case 6:   x=-x;  y=-y;         break;   // --+
                case 7:   x=-x;  y=-y;  z=-z;  break;   // ---
                default:
                    printf("bad case cubic isym_sign in euclidean_op \n");
                    exit(0);
                break;
            } // end switch on isym_sign
        #endif

        #if ALGORITHM_NUMBER==383
            switch(isym%24) {
                case 0 :   x=(*q).x;  y=(*q).y;  z=(*q).z;   w=(*q).w;  break; 
                case 1 :   x=(*q).x;  y=(*q).y;  z=(*q).w;   w=(*q).z;  break; 
                case 2 :   x=(*q).x;  y=(*q).z;  z=(*q).y;   w=(*q).w;  break; 
                case 3 :   x=(*q).x;  y=(*q).z;  z=(*q).w;   w=(*q).y;  break; 
                case 4 :   x=(*q).x;  y=(*q).w;  z=(*q).y;   w=(*q).z;  break; 
                case 5 :   x=(*q).x;  y=(*q).w;  z=(*q).z;   w=(*q).y;  break; 

                case 6 :   x=(*q).y;  y=(*q).x;  z=(*q).w;   w=(*q).z;  break; 
                case 7 :   x=(*q).y;  y=(*q).x;  z=(*q).z;   w=(*q).w;  break; 
                case 8 :   x=(*q).y;  y=(*q).z;  z=(*q).x;   w=(*q).w;  break; 
                case 9 :   x=(*q).y;  y=(*q).z;  z=(*q).w;   w=(*q).x;  break; 
                case 10:   x=(*q).y;  y=(*q).w;  z=(*q).x;   w=(*q).z;  break; 
                case 11:   x=(*q).y;  y=(*q).w;  z=(*q).z;   w=(*q).x;  break; 

                case 12:   x=(*q).z;  y=(*q).x;  z=(*q).y;   w=(*q).w;  break; 
                case 13:   x=(*q).z;  y=(*q).x;  z=(*q).w;   w=(*q).y;  break; 
                case 14:   x=(*q).z;  y=(*q).y;  z=(*q).x;   w=(*q).w;  break; 
                case 15:   x=(*q).z;  y=(*q).y;  z=(*q).w;   w=(*q).x;  break; 
                case 16:   x=(*q).z;  y=(*q).w;  z=(*q).x;   w=(*q).y;  break; 
                case 17:   x=(*q).z;  y=(*q).w;  z=(*q).y;   w=(*q).x;  break; 

                case 18:   x=(*q).w;  y=(*q).x;  z=(*q).y;   w=(*q).z;  break; 
                case 19:   x=(*q).w;  y=(*q).x;  z=(*q).z;   w=(*q).y;  break; 
                case 20:   x=(*q).w;  y=(*q).y;  z=(*q).x;   w=(*q).z;  break; 
                case 21:   x=(*q).w;  y=(*q).y;  z=(*q).z;   w=(*q).x;  break; 
                case 22:   x=(*q).w;  y=(*q).z;  z=(*q).x;   w=(*q).y;  break; 
                case 23:   x=(*q).w;  y=(*q).z;  z=(*q).y;   w=(*q).x;  break; 
                default:  
                    printf("bad case cubic isym_perm in euclidean_op \n");
                    exit(0);
                break;
            } // end switch on isym_perm
            switch(isym/24) {
                case 0:                              break; 
                case 1:                        w=-w; break; 
                case 2:                 z=-z;        break; 
                case 3:                 z=-z;  w=-w; break; 
                case 4:          y=-y;               break; 
                case 5:          y=-y;         w=-w; break; 
                case 6:          y=-y;  z=-z;        break; 
                case 7:          y=-y;  z=-z;  w=-w; break; 
                case 8:   x=-x;                      break; 
                case 9:   x=-x;                w=-w; break; 
                case 10:  x=-x;         z=-z;        break; 
                case 11:  x=-x;         z=-z;  w=-w; break; 
                case 12:  x=-x;  y=-y;               break; 
                case 13:  x=-x;  y=-y;         w=-w; break; 
                case 14:  x=-x;  y=-y;  z=-z;        break; 
                case 15:  x=-x;  y=-y;  z=-z;  w=-w; break; 
                default:
                    printf("bad case hyper isym_sign in euclidean_op \n");
                    exit(0);
                break;
            } // end switch on isym_sign
        #endif

        // Following for perc, ising are to avoid compiler warning
        #ifdef PERCOLATION_HEXAGONAL_LATTICE
            // identity
            x=(*q).x;      y=(*q).y;   
        #endif

        #ifdef ISING_SQUARE_LATTICE
            x=(*q).x;  y=(*q).y; // identity
        #endif
        #ifdef ISING_TRIANGULAR_LATTICE
            x=(*q).x;  y=(*q).y; // identity
        #endif


        x+=(*p).x;
        y+=(*p).y;
        #ifdef THREE_DIMENSIONS
            z+=(*p).z;
        #endif
        #ifdef FOUR_DIMENSIONS
             z+=(*p).z;
             w+=(*p).w;
        #endif
    } // end euclidean_op()


    inline long coord_x() {return(x);} 
    inline long coord_y() {return(y);} 
    #ifdef THREE_DIMENSIONS
        inline long coord_z() {return(z);} 
    #endif
    #ifdef FOUR_DIMENSIONS
        inline long coord_z() {return(z);} 
        inline long coord_w() {return(w);} 
    #endif
    long coord_x_slow();
    long coord_y_slow();
    double real_coord_x();
    double real_coord_y();
    #ifdef TWO_DIMENSIONS
        void assign(long xx,long yy);
    #endif
    #ifdef THREE_DIMENSIONS
        long coord_z_slow();
        double real_coord_z();
        void assign(long xx,long yy,long zz);
    #endif
    #ifdef FOUR_DIMENSIONS
        long coord_z_slow();
        double real_coord_z();
        long coord_w_slow();
        double real_coord_w();
        void assign(long xx,long yy,long zz,long ww);
    #endif

    inline long excluded_distance(excluded_class* excluded);

    // Ising routines
    long point_to_index(long nx, long ny)
    {
        #ifdef ISING_SQUARE_LATTICE
            // test for being out of bounds ??
            return(x+y*nx);
        #endif
        #ifdef ISING_TRIANGULAR_LATTICE
            long i,j;
            if ((x+y)%3!=2) return(-1);
            // x-y can be negative, hence the shifts
            i=(x-y+10*nx)/2-5*nx;
            j=(x+y-2)/3;
            if (j<0 || j>=ny || i<0 || i>=nx) return(-1);
            return(i+j*nx);
        #endif

        printf("error in ising::point_to_index()\n");
        exit(10);
    }

    long boundary(long nx, long ny, long ising_bc)
/*
Returns
     0 for interior pts, 
     +/- 1 for fixed boundary spin 
     +2 free boundary spin  
     +3 antiperiodic boundary spin
     +4 periodic boundary spin

ising_bc==1; Dobrushin

     +++++++++----------
     ___________________
   +|                   |-
   +|                   |-
   +|                   |-
   +|                   |-
   +|                   |-
   +|___________________|-
     +++++++++----------


ising_bc==2; Dobrushin, antiperiodic

     +++++++++----------
     ___________________
    |                   | 
   a|                   |a 
   n|                   |n 
   t|                   |t 
   i|                   |i 
    |___________________|
     +++++++++----------


ising_bc==3; Dobrushin/free, antiperiodic

     ______free_________
    |                   | 
   a|                   |a 
   n|                   |n 
   t|                   |t 
   i|                   |i 
    |___________________|
     +++++++++----------

ising_bc==4; Dobrushin except free on top

     _______free________
   +|                   |-
   +|                   |-
   +|                   |-
   +|                   |-
   +|                   |-
   +|___________________|-
     +++++++++----------

ising_bc==5; +1 all around

     +++++++++++++++++++
     ___________________ 
   +|                   |+
   +|                   |+
   +|                   |+
   +|                   |+
   +|                   |+
   +|___________________|+
     +++++++++++++++++++

*/
    {
        #ifdef ISING_SQUARE_LATTICE
            if (x>0 && x<nx-1 && y>0 && y<ny-1) {
                return(0);
            }

            switch(ising_bc) {
                case 1: // Dobrushin
                    if (x<nx/2) return(-1);
                    else return(1);
                break;
                case 2: // Dobrushin, antiperiodic
                    if (x==0 && 0<y && y<ny-1) return(3); // left: antiper.
                    if (x==nx-1 && 0<y && y<ny-1) return(3); // right: antiper.
                    if (y==0 && x < nx/2) return(-1); // bottom left : -
                    if (y==0 && x >= nx/2) return(+1); // bottom right : +
                    if (y==ny-1 && x < nx/2) return(-1); // top left : -
                    if (y==ny-1 && x >= nx/2) return(+1); // top right : +
                break;
                case 3: // Dobrushin/free, antiperiodic
                    // NB : top corners are in left/right sides
                    if (x==0 && 0<y && y<=ny-1) return(3); // left : antiper.
                    if (x==nx-1 && 0<y && y<=ny-1) return(3);// right: antiper.
                    if (y==0 && x < nx/2) return(-1); // bottom left : - 
                    if (y==0 && x >= nx/2) return(+1); // bottom right : +
                    if (y==ny-1 && x < nx/2) return(2); // top left : free
                    if (y==ny-1 && x >= nx/2) return(2); // top right : free
                break;
	    }
        #endif
        #ifdef ISING_TRIANGULAR_LATTICE
            long i,j;
            if ((x+y)%3!=2) return(-2);
            // x-y can be negative, hence the shifts
            i=(x-y+10*nx)/2-5*nx;
            j=(x+y-2)/3;
            if (i>0 && i<nx-1 && j>0 && j<ny-1) {
                return(0);
            } 
            switch(ising_bc) {
                case 0: // free on all sides
                    return(2);
                break;
                case 1: // Dobrushin
                    if (i<nx/2) return(-1);
                    else return(1);
                break;
                case 2: // Dobrushin, antiperiodic
                    if (i==0 && 0<j && j<ny-1) return(3); // left: antiper.
                    if (i==nx-1 && 0<j && j<ny-1) return(3); // right: antiper.
                    if (j==0 && i < nx/2) return(-1); // bottom left : -
                    if (j==0 && i >= nx/2) return(+1); // bottom right : +
                    if (j==ny-1 && i < nx/2) return(-1); // top left : -
                    if (j==ny-1 && i >= nx/2) return(+1); // top right : +
                break;
                case 3: // Dobrushin/free, antiperiodic
                    // NB : top corners are in left/right sides
                    if (i==0 && 0<j && j<=ny-1) return(3); // left : antiper.
                    if (i==nx-1 && 0<j && j<=ny-1) return(3); // right:antiper.
                    if (j==0 && i < nx/2) return(-1); // bottom left : - 
                    if (j==0 && i >= nx/2) return(+1); // bottom right : +
                    if (j==ny-1 && i < nx/2) return(2); // top left : free
                    if (j==ny-1 && i >= nx/2) return(2); // top right : free
                break;
                case 4: // Dobrushin except for free on top
                    // NB : top corners are in left/right sides
                    if (i==0 && 0<j && j<=ny-1) return(-1); // left : -
                    if (i==nx-1 && 0<j && j<=ny-1) return(+1); // right: +
                    if (j==0 && i < nx/2) return(-1); // bottom left : - 
                    if (j==0 && i >= nx/2) return(+1); // bottom right : +
                    if (j==ny-1 && i < nx/2) return(2); // top left : free
                    if (j==ny-1 && i >= nx/2) return(2); // top right : free
                break;
                case 5: // + on all sides
                    return(1);
                break;
	    }
        #endif

        printf("error in boundary()\n");
        exit(10);
    }

};  // end class point

inline point operator+(point p,point q)
{
    point r=p; 
    return r+=q;
} 

inline point operator-(point p,point q)
{
    point r=p; 
    return r-=q;
} 

int operator==(point p,point q);
point operator*(long c,point p);
long operator*(point p,point q);
void htab_clear(point *htab_point);

long operator*(point p,point q);

#endif

///////////////////////////////////////////////////////////////
//  checkk checkk checkk checkk checkk checkk checkk checkk  //
///////////////////////////////////////////////////////////////
