/*
    SAW_pivot implements the pivot algorithm for Monte Carlo simulations
      of self-avoiding walks
    Copyright (C) 2003 Tom Kennedy

    This file is part of SAW_pivot (version 1.0).

    SAW_pivot is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SAW_pivot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SAW_pivot; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    E-mail contact: tgk@math.arizona.edu
    Ordinary mail: 
        Tom Kennedy
        Mathematics Department
        University of Arizona
        Tucson, ZA 85721, USA
*/

// file name : src/rpoint.c

#include <cstdio> 
using namespace std; 
#include "math.h"
#include <iostream>
using namespace std; 
#include <cstdlib>
#include "sys/time.h"
#include "time.h"
#include "limits.h"
#include <complex>
using namespace std;

#define I complex<double>(0.,1.);

#include "local_constants.h"
#include "point.h"
#include "rpoint.h"
#include "lib.h"

extern long group_product[NUM_SYM+1][NUM_SYM+1];
extern long group_inverse[NUM_SYM+1];

////////////////////////////////////////////////////////////////
//             member functions for class rpoint               //
////////////////////////////////////////////////////////////////

void rpoint::print(FILE *fptr) 
{
    #ifdef TWO_DIMENSIONS
        fprintf(fptr,"%f %f \n",x,y);
    #endif
    #ifdef THREE_DIMENSIONS
        fprintf(fptr,"%f %f %f\n",x,y,z);
    #endif
    #ifdef FOUR_DIMENSIONS
        fprintf(fptr,"%f %f %f %f\n",x,y,z,w);
    #endif
}

void rpoint::scanf(FILE *fptr) 
{
    #ifdef TWO_DIMENSIONS
        fscanf(fptr,"%lf %lf\n",&x,&y);
    #endif
    #ifdef THREE_DIMENSIONS
        fscanf(fptr,"%lf %lf %lf\n",&x,&y,&z);
    #endif
    #ifdef FOUR_DIMENSIONS
        fscanf(fptr,"%lf %lf %lf %lf\n",&x,&y,&z,&w);
    #endif
}

void rpoint::write(FILE *fptr) 
{
    fwrite(&x,sizeof(double),1,fptr);
    fwrite(&y,sizeof(double),1,fptr);
    #ifdef THREE_DIMENSIONS
        fwrite(&z,sizeof(double),1,fptr);
    #endif
    #ifdef FOUR_DIMENSIONS
        fwrite(&z,sizeof(double),1,fptr);
        fwrite(&w,sizeof(double),1,fptr);
    #endif
}

void rpoint::read(FILE *fptr)
{
    fread(&x,sizeof(double),1,fptr);
    fread(&y,sizeof(double),1,fptr);
    #ifdef THREE_DIMENSIONS
        fread(&z,sizeof(double),1,fptr);
    #endif
    #ifdef FOUR_DIMENSIONS
        fread(&z,sizeof(double),1,fptr);
        fread(&w,sizeof(double),1,fptr);
    #endif
}

rpoint& rpoint::operator/=(double c) 
{
    x/=c;
    y/=c;
    #ifdef THREE_DIMENSIONS
        z/=c;
    #endif
    #ifdef FOUR_DIMENSIONS
        z/=c;
        w/=c;
    #endif
    return *this;
} 

rpoint& rpoint::operator*=(double c) 
{
    x*=c;
    y*=c;
    #ifdef THREE_DIMENSIONS
        z*=c;
    #endif
    #ifdef FOUR_DIMENSIONS
        z*=c;
        w*=c;
    #endif
    return *this;
} 

void rpoint::zero() 
{
    x=0.; y=0.;
    #ifdef THREE_DIMENSIONS
        z=0.;
    #endif
    #ifdef FOUR_DIMENSIONS
        z=0.;
        w=0.;
    #endif
} 


#ifdef TWO_DIMENSIONS
void rpoint::assign(double xx,double yy) 
{
    x=xx; y=yy;
} 
#endif

#ifdef THREE_DIMENSIONS
void rpoint::assign(double xx,double yy,double zz) 
{
    x=xx; y=yy; z=zz; 
} 
#endif

#ifdef FOUR_DIMENSIONS
void rpoint::assign(double xx,double yy,double zz,double ww) 
{
    x=xx; y=yy; z=zz; w=ww;
} 
#endif

double rpoint::coord_x() 
{
    return(x);
}

double rpoint::coord_y() 
{
    return(y);
}

#ifdef THREE_DIMENSIONS
double rpoint::coord_z() 
{
    return(z);
}
#endif

#ifdef FOUR_DIMENSIONS
double rpoint::coord_z() 
{
    return(z);
}

double rpoint::coord_w() 
{
    return(w);
}
#endif

rpoint operator*(double c,rpoint p) 
{
    rpoint r=p; 
    return r*=c;
} 

rpoint operator*(rpoint p ,double c)
{
    rpoint r=p; 
    return r*=c;
} 

rpoint operator/(rpoint p ,double c)
{
    rpoint r=p; 
    return r/=c;
} 

rpoint& rpoint::operator=(point p)  
{
    x=p.real_coord_x(); 
    y=p.real_coord_y(); 
    #ifdef THREE_DIMENSIONS
        z=p.real_coord_z(); 
    #endif
    #ifdef FOUR_DIMENSIONS
        z=p.real_coord_z(); 
        w=p.real_coord_w(); 
    #endif
    return(*this);
} // end point::operator=()


#ifdef TWO_DIMENSIONS
void rpoint::rotate(double theta) 
{
    double xx,yy,c,s;
    xx=x; yy=y;
    c=cos(theta);
    s=sin(theta);
    x=c*xx-s*yy;
    y=s*xx+c*yy;
} 
#endif

double rpoint::distance()  
// returns euclidean distance from point to origin
{
    #ifdef TWO_DIMENSIONS
        return(sqrt(x*x+y*y));
    #endif
    #ifdef THREE_DIMENSIONS
        return(sqrt(x*x+y*y+z*z));
    #endif
    #ifdef FOUR_DIMENSIONS
        return(sqrt(x*x+y*y+z*z+w*w));
    #endif
} // end rpoint::distance()

double rpoint::distance_projection(double a, double b)  
// projects point onto plane spanned by y axis and (a,b), where (a,b) has 
// length 1, then returns euclidean distance from projection to origin
{
    #ifdef THREE_DIMENSIONS
        double c=x*a+z*b;
        return(sqrt(c*c+y*y));
    #endif
    printf("error: distance_projection called, not in 3 dimensions\n"); 
    exit(0);
    return(0.);
} // end rpoint::distance_projection() 

double rpoint::p_angle() 
{
    #ifdef TWO_DIMENSIONS
        double angle=polar_angle(x,y);
        return(angle);
    #endif
    printf("\nERROR: rpoint::p_angle() called in dimension other than 2\n");
    exit(0);
    return(0.); // to avoid compiler warning 
} // end rpoint::p_angle() 

///////////////////////////////////////////////////////////////
//  checkk checkk checkk checkk checkk checkk checkk checkk  //
///////////////////////////////////////////////////////////////
