/*
    SAW_pivot implements the pivot algorithm for Monte Carlo simulations
      of self-avoiding walks
    Copyright (C) 2003 Tom Kennedy

    This file is part of SAW_pivot (version 1.0).

    SAW_pivot is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SAW_pivot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SAW_pivot; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    E-mail contact: tgk@math.arizona.edu
    Ordinary mail:
        Tom Kennedy
        Mathematics Department
        University of Arizona
        Tucson, ZA 85721, USA
*/

// file name : src/walk.c
//

// system includes
#include <cstdio>
using namespace std;
#include "math.h"
#include <iostream>
using namespace std;
#include <cstdlib>
#include "sys/time.h"
#include "time.h"
#include "limits.h"
#include <complex>
using namespace std;

#define I complex<double>(0.,1.);

// local includes
#include "random_number_generator.h"
#include "local_constants.h"
#include "lib.h"
#include "point.h"
#include "rpoint.h"
#include "excluded_class.h"
#include "rwalk.h"
#include "walk.h"

long group_product[NUM_SYM+1][NUM_SYM+1];
long group_inverse[NUM_SYM+1];

void setup_symmetry()
// this routine computes group_product[][] and group_inverse[] entries
// by trial and error
{
    long i,j,k;
    point origin,p1,p2,q1,q2,temp;
    #ifdef THREE_DIMENSIONS
        point p3,q3;
    #endif
    #ifdef FOUR_DIMENSIONS
        point p3,q3,p4,q4;
    #endif

    origin.zero();

    for (i=0;i<=NUM_SYM;i++) for (j=0;j<=NUM_SYM;j++) {
        group_product[i][j]=-1;
        // p1,p2,... = two group operations applied to (1,0),(0,1),...
        #ifdef TWO_DIMENSIONS
            p1.assign(0,1);
            temp=p1; p1.euclidean_op(&temp,&origin,j);
            temp=p1; p1.euclidean_op(&temp,&origin,i);
            p2.assign(1,0);
            temp=p2; p2.euclidean_op(&temp,&origin,j);
            temp=p2; p2.euclidean_op(&temp,&origin,i);
        #endif
        #ifdef THREE_DIMENSIONS
            p1.assign(0,0,1);
            temp=p1; p1.euclidean_op(&temp,&origin,j);
            temp=p1; p1.euclidean_op(&temp,&origin,i);
            p2.assign(0,1,0);
            temp=p2; p2.euclidean_op(&temp,&origin,j);
            temp=p2; p2.euclidean_op(&temp,&origin,i);
            p3.assign(1,0,0);
            temp=p3; p3.euclidean_op(&temp,&origin,j);
            temp=p3; p3.euclidean_op(&temp,&origin,i);
        #endif
        #ifdef FOUR_DIMENSIONS
            p1.assign(0,0,0,1);
            temp=p1; p1.euclidean_op(&temp,&origin,j);
            temp=p1; p1.euclidean_op(&temp,&origin,i);
            p2.assign(0,0,1,0);
            temp=p2; p2.euclidean_op(&temp,&origin,j);
            temp=p2; p2.euclidean_op(&temp,&origin,i);
            p3.assign(0,1,0,0);
            temp=p3; p3.euclidean_op(&temp,&origin,j);
            temp=p3; p3.euclidean_op(&temp,&origin,i);
            p4.assign(1,0,0,0);
            temp=p4; p4.euclidean_op(&temp,&origin,j);
            temp=p4; p4.euclidean_op(&temp,&origin,i);
        #endif
        // search for one group operation which gives same thing
        for (k=0;k<=NUM_SYM;k++) {
            #ifdef TWO_DIMENSIONS
                q1.assign(0,1);
                temp=q1; q1.euclidean_op(&temp,&origin,k);
                q2.assign(1,0);
                temp=q2; q2.euclidean_op(&temp,&origin,k);
                if (p1==q1 && p2==q2) group_product[i][j]=k;
            #endif
            #ifdef THREE_DIMENSIONS
                q1.assign(0,0,1);
                temp=q1; q1.euclidean_op(&temp,&origin,k);
                q2.assign(0,1,0);
                temp=q2; q2.euclidean_op(&temp,&origin,k);
                q3.assign(1,0,0);
                temp=q3; q3.euclidean_op(&temp,&origin,k);
                if (p1==q1 && p2==q2 && p3==q3) group_product[i][j]=k;
            #endif
            #ifdef FOUR_DIMENSIONS
                q1.assign(0,0,0,1);
                temp=q1; q1.euclidean_op(&temp,&origin,k);
                q2.assign(0,0,1,0);
                temp=q2; q2.euclidean_op(&temp,&origin,k);
                q3.assign(0,1,0,0);
                temp=q3; q3.euclidean_op(&temp,&origin,k);
                q4.assign(1,0,0,0);
                temp=q4; q4.euclidean_op(&temp,&origin,k);
                if (p1==q1 && p2==q2 && p3==q3 && p4==q4)
                    group_product[i][j]=k;
            #endif
        }
        if (group_product[i][j]==-1) {
            printf("ERROR in setup_symmetry() %ld %ld\n",i,j);
            exit(1);
        }
    } // end loops on i and j

    for (i=0;i<=NUM_SYM;i++) {
        group_inverse[i]=-1;
        for (k=0;k<=NUM_SYM;k++) {
            #ifdef TWO_DIMENSIONS
                p1.assign(0,1);
                temp=p1; p1.euclidean_op(&temp,&origin,i);
                temp=p1; p1.euclidean_op(&temp,&origin,k);
                q1.assign(0,1);
                p2.assign(1,0);
                temp=p2; p2.euclidean_op(&temp,&origin,i);
                temp=p2; p2.euclidean_op(&temp,&origin,k);
                q2.assign(1,0);
                if (p1==q1 && p2==q2) group_inverse[i]=k;
            #endif
            #ifdef THREE_DIMENSIONS
                p1.assign(0,0,1);
                temp=p1; p1.euclidean_op(&temp,&origin,i);
                temp=p1; p1.euclidean_op(&temp,&origin,k);
                q1.assign(0,0,1);
                p2.assign(0,1,0);
                temp=p2; p2.euclidean_op(&temp,&origin,i);
                temp=p2; p2.euclidean_op(&temp,&origin,k);
                q2.assign(0,1,0);
                p3.assign(1,0,0);
                temp=p3; p3.euclidean_op(&temp,&origin,i);
                temp=p3; p3.euclidean_op(&temp,&origin,k);
                q3.assign(1,0,0);
                if (p1==q1 && p2==q2 && p3==q3) group_inverse[i]=k;
            #endif
            #ifdef FOUR_DIMENSIONS
                p1.assign(0,0,0,1);
                temp=p1; p1.euclidean_op(&temp,&origin,i);
                temp=p1; p1.euclidean_op(&temp,&origin,k);
                q1.assign(0,0,0,1);
                p2.assign(0,0,1,0);
                temp=p2; p2.euclidean_op(&temp,&origin,i);
                temp=p2; p2.euclidean_op(&temp,&origin,k);
                q2.assign(0,0,1,0);
                p3.assign(0,1,0,0);
                temp=p3; p3.euclidean_op(&temp,&origin,i);
                temp=p3; p3.euclidean_op(&temp,&origin,k);
                q3.assign(0,1,0,0);
                p4.assign(0,1,0,0);
                temp=p4; p4.euclidean_op(&temp,&origin,i);
                temp=p4; p4.euclidean_op(&temp,&origin,k);
                q4.assign(0,1,0,0);
                if (p1==q1 && p2==q2 && p3==q3 && p4==q4) group_inverse[i]=k;
            #endif
        }
        if (group_inverse[i]==-1) {
            printf("ERROR in setup_symmetry()\n");
            exit(1);
        }
    } // end loop on i
    return;

} // end setup_symmetry()

////////////////////////////////////////////////////////////////
//             member functions for class walk                //
////////////////////////////////////////////////////////////////

long walk::get_nsteps() {return(nsteps);}
void walk::set_nsteps(long n) {nsteps=n;}
long walk::get_niter() {return(niter);}
void walk::increment_niter() {niter++;}
long walk::get_npivot() {return(npivot);}
void walk::set_npivot(long n) {npivot=n;}

point walk::step_rval(long i)
// This routine computes the ith point on the walk from the complicated
// data structure. Often we duplicate this code for speed rather than call this
{
    point temp;
    long iseg;
    iseg=find_segment(i,npivot,ptime);
    temp.euclidean_op(steps+i,shift+iseg,igroup[iseg]);
    return(temp);
}

void walk::assign_point(long i,point p) {
    if (npivot>0 || igroup[0]!=0) {
        printf("err:walk::assign_point called but npivot>0 or igroup[0]!=0\n");
        exit(1);
    }
    steps[i]=p;
}

void walk::initialize()
// NB this does not create a walk. It only initializes things used by the
// implicit pivot data structure
{
    npivot=0;
    igroup[0]=0;
    shift[0].zero();
    // segment number iseg corresponds to times [ptime[iseg],ptime[iseg+1])
    // So when npivot=0 we should have :
    ptime[0]=0;
    ptime[1]=nsteps+1;
}

void walk::line_initialize(int direction)
{
// initializes walk to be straight line
// 1 is horizontal
// 2 is 45 degs
// 3 is vertical
// 4 is 60 degs
    long i,i1,i2,i3,i4;

    initialize();
    niter=0;
    i1=0; i2=0; i3=0; i4=0;
    for (i=0;i<=nsteps;i++) {
        #ifdef TWO_DIMENSIONS
            steps[i].assign(i1,i2);
        #endif
        #ifdef THREE_DIMENSIONS
            steps[i].assign(i1,i2,i3);
        #endif
        #ifdef FOUR_DIMENSIONS
            steps[i].assign(i1,i2,i3,i4);
        #endif

        #ifdef SQUARE_LATTICE
            switch(direction) {
                case 1: i1++; break;
                case 2:
                    switch (i%2) {
                        case 0: i1++; break;
                        case 1: i2++; break;
                    }
                break;
                case 3:
                    i2++;
                break;
                default:
                    printf("bad case: line_initialize() a\n");
                    exit(0);
                    break;
            }
        #endif

        #if ALGORITHM_NUMBER==5
            switch(direction) {
                case 1:
                    switch (i%4) {
                        case 0: i1++; break;
                        case 1: i1++; i2++; break;
                        case 2: i1++; break;
                        case 3: i2--; break;
                    }
                break;
                case 3:
                    switch (i%2) {
                        case 0: i2++; break;
                        case 1: i1++; i2++; break;
                    }
                break;
                default:
                    printf("bad case line_initialize() b\n");
                    exit(0);
                break;
            }
        #endif

        #if ALGORITHM_NUMBER==11
            switch(direction) {
                case 1:  i1++; break;
                case 3:
                    switch (i%2) {
                        case 0: i2++; break;
                        case 1: i1++; i2++; break;
                    }
                break;
                case 4:
                    i1++; i2++;
                break;
                default:
                    printf("bad case line_initialize() c\n");
                    exit(0);
                break;
            }
        #endif

        #ifdef CUBIC_LATTICE
            switch(direction) {
                case 1: i1++; break;
                case 2: if (i%2==0) i2++; else i1++; break;
                case 3: i2++; break;
                default:
                    printf("bad case line_initialize() d\n");
                    exit(0);
                break;
            }
        #endif

        #ifdef HYPER_LATTICE
            switch(direction) {
                case 1: i1++; break;
                case 2: if (i%2==0) i2++; else i1++; break;
                case 3: i2++; break;
                default:
                    printf("bad case line_initialize() e\n");
                    exit(0);
                break;
            }
        #endif

    }
} // end line_initialize()

#ifdef TWO_DIMENSIONS
void walk::two_line_initialize(int direction)
{
// initializes two walks to be straight lines
// 10 lines are at angles +/- angles
    long i,i1,i2;

    initialize();
    niter=0;

    // line at 45
    i1=0; i2=1;
    for (i=nsteps/2;i<=nsteps;i++) {
        steps[i].assign(i1,i2);
        if (i%2==0) i1++;
        else i2++;
    }

    // line at -45
    i1=0; i2=1;
    for (i=nsteps/2;i>=0;i--) {
        steps[i].assign(i1,i2);
        if (i%2==0) i1--;
        else i2++;
    }

} // end two_line_initialize()
#endif

int walk::allocate(long n)
{
    steps = new point[n+1];
    // this may be one larger than needed
    ptime = new long[MAX_NPIVOT+2];
    igroup = new long[MAX_NPIVOT+1];
    shift = new point[MAX_NPIVOT+1];
    if (steps==0 || ptime==0 || igroup==0 || shift==0) { //allocation failed
        printf("allocate error in walk::allocate \n");
        return(0);
    }
    nsteps=n;
    return(1);
}

void walk::deallocate()
{
    delete [] steps;
    delete [] ptime;
    delete [] igroup;
    delete [] shift;
};

void walk::print(FILE *fptr)
// this prints the coordinates of the walk wrt the two lattice vectors
// not the actual x,y coords
// If npivot is > 0, it also prints the arrays ptime, igroup, shift.
// This is for debugging. It is incompatible with walk::scanf()
{
    long i;
    fprintf(fptr,"%ld %ld\n",nsteps,niter);
    for (i=0;i<=nsteps;i++) steps[i].print(fptr);
    if (npivot>0 || igroup[0]!=0) {
        fprintf(fptr,"npivot=%ld\n",npivot);
        fprintf(fptr,"iseg     ptime   igroup    shift \n");
        for (i=0;i<=npivot;i++) {
            fprintf(fptr,"%4ld  %8ld   %2ld    ",i,ptime[i],igroup[i]);
                shift[i].print(fptr);
        }
    }
}

void walk::scan(FILE *fptr)
{
    long i;
    fscanf(fptr,"%ld %ld",&nsteps,&niter);
    for (i=0;i<=nsteps;i++) steps[i].scan(fptr);
    initialize();
}

void walk::plot(FILE *fptr)
// this prints out the actual x,y,z,... coordinates of the walk so it
// can be plotted
{
    long i;
    if (npivot!=0) {
        printf("ERROR: walk::print() called with npivot!=0 \n");
        exit(1L);
    }

    for (i=0;i<=nsteps;i++) {
        #ifdef TWO_DIMENSIONS
            fprintf(fptr,"%f %f \n",steps[i].real_coord_x(),
                steps[i].real_coord_y());
        #endif
        #ifdef THREE_DIMENSIONS
            fprintf(fptr,"%f %f %f\n",steps[i].real_coord_x(),
                steps[i].real_coord_y(),steps[i].real_coord_z());
        #endif
        #ifdef FOUR_DIMENSIONS
            fprintf(fptr,"%f %f %f %f\n",steps[i].real_coord_x(),
                steps[i].real_coord_y(),steps[i].real_coord_z(),
                steps[i].real_coord_w());
        #endif
    }
} // end walk::plot()

void walk::iplot(FILE *fptr)
// this prints out integer coordinates of the walk, not the actual x,y,z,...
{
    long i;
    if (npivot!=0) {
        printf("ERROR: walk::iplot() called with npivot!=0 \n");
        exit(1L);
    }

    for (i=0;i<=nsteps;i++) {
        #ifdef TWO_DIMENSIONS
            fprintf(fptr,"%ld %ld \n",steps[i].coord_x(),
                steps[i].coord_y());
        #endif
        #ifdef THREE_DIMENSIONS
            fprintf(fptr,"%ld %ld %ld\n",steps[i].coord_x(),
                steps[i].coord_y(),steps[i].coord_z());
        #endif
        #ifdef FOUR_DIMENSIONS
            fprintf(fptr,"%ld %ld %ld %ld\n",steps[i].coord_x(),
                steps[i].coord_y(),steps[i].coord_z(),
                steps[i].coord_w());
        #endif
    }
} // end walk::iplot()

void walk::bi_plot(FILE *fptr1,FILE *fptr2,long nseg)
// divides walk into two sets of segments; dstep is number of points in segment
{
    long i,j,dstep;
    dstep=nsteps/nseg;
    if (dstep==0) return;
    if (npivot!=0) {
        printf("ERROR: walk::print() called with npivot!=0 \n");
        exit(1L);
    }

    i=0;
    for (;;) {
        for (j=0;j<dstep;j++) {
            #ifdef TWO_DIMENSIONS
                fprintf(fptr1,"%f %f \n",steps[i].real_coord_x(),
                    steps[i].real_coord_y());
            #endif
            #ifdef THREE_DIMENSIONS
                fprintf(fptr1,"%f %f %f\n",steps[i].real_coord_x(),
                    steps[i].real_coord_y(),steps[i].real_coord_z());
            #endif
            #ifdef FOUR_DIMENSIONS
                fprintf(fptr1,"%f %f %f %f\n",steps[i].real_coord_x(),
                    steps[i].real_coord_y(),steps[i].real_coord_z(),
                    steps[i].real_coord_w());
            #endif
            i++;
            if (i>nsteps) return;
	}
        fprintf(fptr1,"\n");

        for (j=0;j<dstep;j++) {
            #ifdef TWO_DIMENSIONS
                fprintf(fptr2,"%f %f \n",steps[i].real_coord_x(),
                    steps[i].real_coord_y());
            #endif
            #ifdef THREE_DIMENSIONS
                fprintf(fptr2,"%f %f %f\n",steps[i].real_coord_x(),
                    steps[i].real_coord_y(),steps[i].real_coord_z());
            #endif
            #ifdef FOUR_DIMENSIONS
                fprintf(fptr2,"%f %f %f %f\n",steps[i].real_coord_x(),
                    steps[i].real_coord_y(),steps[i].real_coord_z(),
                    steps[i].real_coord_w());
            #endif
            i++;
            if (i>nsteps) return;
	}
        fprintf(fptr2,"\n");
    }
} // end walk::plot()

void walk::plot_projected(FILE *fptr,double rrho)
// this prints out the x,y coordinates of the walk projected using rrho
{
    long i;
    double x,y;
    rpoint rpt;

    if (npivot!=0) {
        printf("ERROR: walk::print() called with npivot!=0 \n");
        exit(1L);
    }

    #ifdef THREE_DIMENSIONS
        double a=cos(M_PI*rrho/180);
        double b=sin(M_PI*rrho/180);
    #endif

    for (i=0;i<=nsteps;i++) {
        // compute x,y
        rpt=steps[i];
        #ifdef TWO_DIMENSIONS
            x=rpt.coord_x();
        # endif
        #ifdef THREE_DIMENSIONS
            x=a*rpt.coord_x()+b*rpt.coord_z();
        # endif
        y=rpt.coord_y();
        fprintf(fptr,"%f %f\n",x,y);
    } // end loop on i
} // end walk::plot_projected()

walk& walk::operator=(walk& w)
{
    long i;
    if (w.get_npivot()!=0) {
        printf("ERROR in walk::=, npivot!=0 \n");
        exit(1);
    }
    (*this).initialize();
    nsteps=w.get_nsteps();
    niter=w.get_niter();
    for (i=0;i<=nsteps;i++) steps[i]=w.steps[i];
    return(*this);
}

void walk::test(excluded_class* excluded)
{
// this tests that the walk is a n.n. walk and that it does not hit the
// excluded region
// If it fails the test we just crash (exit).
    long i,iseg;
    point stepsi;

    // n.n. check
    for (i=1;i<=nsteps;i++)
        if (fabs((steps[i-1]-steps[i]).distance()-1.) > 1.e-5) {
            printf("BAD WALK  BAD WALK   BAD WALK   BAD WALK   BAD WALK \n");
            printf("BAD WALK  BAD WALK   BAD WALK   BAD WALK   BAD WALK \n");
            printf("not nearest neighbor\n");
            exit(1);
        }

    for (i=1;i<=nsteps;) {
        // stepsi is w[i]
        iseg=find_segment(i,npivot,ptime);
        stepsi.euclidean_op(steps+i,shift+iseg,igroup[iseg]);
        long separation=stepsi.excluded_distance(excluded);
        if (separation==0) {
            printf("BAD WALK  BAD WALK   BAD WALK   BAD WALK   BAD WALK \n");
            printf("BAD WALK  BAD WALK   BAD WALK   BAD WALK   BAD WALK \n");
            printf("hits excluded region\n");
            exit(1);
        }
        i+=separation;
    } // end loop on i
} // end walk::test()

double walk::turn_frac()
{
// Finds fraction of times at which walk turns.
// For hexagonal lattice we look at next nearest neighbors
    long i,count;
    count=0;

    // hexagonal lattice needs special treatment: n.n.n. always same distance
    #if ALGORITHM_NUMBER==5
        for (i=1;i<nsteps-1;i++)
            if (fabs((steps[i-1]-steps[i+2]).distance()-sqrt(7.)) > 1.e-5)
                count++;
    #else
        for (i=1;i<nsteps;i++)
            if (fabs((steps[i-1]-steps[i+1]).distance()-2.) > 1.e-5) count++;
    #endif

    return(double(count)/double(nsteps-1));
} // end turn_frac

int walk::even(long pivot_loc)
// returns 1 if sum of coords of walks at time pivot is even
// used by Manhattan lattice
{
    #ifdef TWO_DIMENSIONS
        if ((steps[pivot_loc].coord_x()+steps[pivot_loc].coord_y())%2==0)
            return(1);
        else return(0);
    #endif
    #ifdef THREE_DIMENSIONS
        if ((steps[pivot_loc].coord_x()+steps[pivot_loc].coord_y()
            +steps[pivot_loc].coord_z())%2==0) return(1);
        else return(0);
    #endif
    #ifdef FOUR_DIMENSIONS
       if ((steps[pivot_loc].coord_x()+steps[pivot_loc].coord_y()
       +steps[pivot_loc].coord_z()+steps[pivot_loc].coord_w())%2==0) return(1);
       else return(0);
    #endif
} // end walk::even()

inline long point::excluded_distance(excluded_class* excluded)
// Computes "distance" from point to excluded region.
// NB: this routine assumes the point is not inside the excluded region.
//  No error checking is done
// Ideally, distance is the minimum number of nearest neighbor steps needed
// to get from the point to the excluded region, but we actually use
// the l2 distance which is a lower bound on the number of nearest neighbor
// steps. We can round this distance up to the nearest integer. Hence
// the use of ceil() in the above. We use ceil(dist-1.e-10) to be sure
// a number doesn't get wrongly rounded up due to numerical error.
//
// cut plane: the excluded region is only the positive x-axis which may not
// include any points at all. So we widen the excluded region to be the
// strip x>0, -0.5<=y<=0.5. Every bond that crosses the positive x-axis has
// an endpoint in this strip.
{
    double xx,yy,zz;
    zz=0.; // solely to avoid compiler warning
    rpoint rp;
    rp=*this;
    if ((*excluded).region>=0) {
        xx=rp.coord_x();
        yy=rp.coord_y();
    }
    else {
        xx=inner_prod(&rp,&((*excluded).xn));
        yy=inner_prod(&rp,&((*excluded).yn));
    }

    switch (abs((*excluded).region)) {
        case 0:  // no excluded region, full plane
            return(LONG_MAX/2);
        break;
        case 1:  // half plane, walk must satisfy y>0
            if (yy>0.) return(long(ceil(yy-1.e-10)));
            else return(0);
        break;
        case 2:  // cut: walk not allowed to hit positive x-axis
            if (xx<0.) return(long(ceil(sqrt(xx*xx+yy*yy)-1.e-10)));
            else {
                if (fabs(yy)<=0.5) return(0);
                else return(long(ceil(fabs(yy)-1.e-10)));
            }
        break;
        case 3:  // quarter: walk not allowed to hit pos x-axis or pos y-axis
            if (yy>0 && xx>0) {
                if (yy>xx) return(long(ceil(xx-1.e-10)));
                else return(long(ceil(yy-1.e-10)));
            }
            else return(0);
        break;
        case 4: // other quarter plane: walk restricted to z>=0, y>0
            #ifdef THREE_DIMENSIONS
                zz=rp.coord_z();
                if (yy>0 && zz>=0) {
                    //+1 cause exclude is z<=-1
                    if (yy>zz+1) return(long(ceil(zz+1.-1.e-10)));
                    else return(long(ceil(yy-1.e-10)));
                }
                else return(0);
            #endif
                printf("ERROR : excluded=4 and number of dimensions is 2\n");
                exit(1);
        break;
        default:
            printf("excluded case not implemented in excluded_distance \n");
            exit(10);
        break;
    }
    printf("algorithm not implemented in excluded_distance \n");
    exit(1);
} // end point::excluded_distance()


long walk::test_pivot_strictly_saw(
    long pivot_loc,long isym,excluded_class excluded)
// This is the version that changes i and j "simultaneously"
// This routine only checks if the pivot produces a SAW.
// It does not carry out the pivot and it does not check if the new
// walk hits the excluded region.
// Routine returns count if the new walk  is self-avoiding,
// returns -count if the pivot produces a self-intersection,
// where count is the number of distance computations done in this routine.
{
    long i,j,ip,jp,iseg,jseg,igroup_jseg,igroup_iseg,imin,imax,jmin,jmax;
    long separation,min_separation,sep_mod;
    point origin,transi,transj,pp,stepsp,stepsi,stepsj,shift_jseg,shift_iseg;
    long count,changei_flag;

    // pivot is given by w[t] -> g (w[t]-w[pivot_loc])+w[pivot_loc]
    // this is equivalent to w[t] -> g w[t] + trans
    // where trans= w[pivot_loc] - g w[pivot_loc]
    // stepsp=w[pivot_loc]

    origin.zero();
    iseg=find_segment(pivot_loc,npivot,ptime);
    stepsp.euclidean_op(steps+pivot_loc,shift+iseg,igroup[iseg]);
    transi.euclidean_op(&stepsp,&origin,isym);
    transi=stepsp-transi;
    transj.euclidean_op(&stepsp,&origin,group_inverse[isym]);
    transj=stepsp-transj;

    /////////////////////////////////////////////////////////////////////////
    //
    // "Simultaneous" changing of i and j.
    // Given j<pivot_loc<i, we assume w(jp)!=w(ip)
    // for j<jp<pivot_loc<ip<i
    // We then either decrease j or increase i.
    //
    /////////////////////////////////////////////////////////////////////////

    count=0;
    jmin=0; jmax=pivot_loc-1;
    imin=pivot_loc+1; imax=nsteps;

    j=pivot_loc-2;
    i=pivot_loc+1;
    // if pivot_loc==0, j=-2 which would allow jp=-1
    if (j<jmin-1) j=jmin-1;

    if (pivot_loc>nsteps) i=imax+1;

    while (i<=imax || j>=jmin) {
        // changei_flag=1 means we will increase i, =0 means we will increase j
        // We change the index that is closer to pivot_loc (if allowed).
        if (i-pivot_loc>pivot_loc-j) changei_flag=0;
        else changei_flag=1;
        if (i>imax) changei_flag=0;
        if (j<jmin) changei_flag=1;

        if (changei_flag) { // increase i.
	    // Need lower bound on dist from pivoted w[i] to
            // {w[jp]: j<jp<pivot_loc}
            // This lower bound will be min_separation
            // stepsi is w[i]
            iseg=find_segment(i,npivot,ptime);
            stepsi.euclidean_op(steps+i,shift+iseg,igroup[iseg]);
            // pp is w[i] after pivot
            pp.euclidean_op(&stepsi,&transi,isym);
            min_separation=nsteps;
            jseg=npivot;   // can change to jseg=iseg; ?
            shift_jseg=shift[jseg]-pp;
            igroup_jseg=igroup[jseg];
            for (jp=jmax;jp>j;) { // note that j is decreased
                if (ptime[jseg]>jp) {
                    #ifdef USE_FIND_SEGMENT
                        jseg=find_segment(jp,npivot,ptime);
                    #else
                        while (ptime[jseg]>jp) jseg--;
                    #endif
                    shift_jseg=shift[jseg]-pp;
                    igroup_jseg=igroup[jseg];
                }
                // stepsj is w[jp]
                stepsj.euclidean_op(steps+jp,&shift_jseg,igroup_jseg);
                separation=stepsj.lattice_distance();
                count++;
                if (separation==0) return(-count);
                if (separation>=min_separation) {
                    jp-=separation-min_separation+1;
                }
                else {
                    #ifdef USE_THIRD
                        sep_mod=separation%3;
                        min_separation=(2*separation+sep_mod)/3;
                        jp-=1+(separation-sep_mod)/3;
                    #else
                        sep_mod=separation%2;
                        min_separation=(separation+sep_mod)/2;
                        jp-=1+(separation-sep_mod)/2;
                    #endif
                }
            } // end loop on jp
            i+=min_separation;
            if (i>imax+1) i=imax+1;
        } // end increase i

        else {// decrease j.
            //Need lower bound on distance from w[j] to
            // pivoted {w[ip]: pivot_loc<ip<i}
            // Equivalently we can use a lower bound on distance from inverse
            // pivoted w[j] and {w[ip]: pivot_loc<ip<i}
            // This lower bound will be min_separation
            // stepsj is w[j]
            jseg=find_segment(j,npivot,ptime);
            stepsj.euclidean_op(steps+j,shift+jseg,igroup[jseg]);
            pp.euclidean_op(&stepsj,&transj,group_inverse[isym]);
            min_separation=nsteps;
            iseg=0;
            shift_iseg=shift[iseg]-pp;
            igroup_iseg=igroup[iseg];
            for (ip=imin;ip<i;) { // note that i is increased
                if (ptime[iseg+1]<=ip) { // check this
                    #ifdef USE_FIND_SEGMENT
                        iseg=find_segment(ip,npivot,ptime);
                    #else
                        while (ptime[iseg+1]<=ip) iseg++;
                    #endif
                    shift_iseg=shift[iseg]-pp;
                    igroup_iseg=igroup[iseg];
                }
                stepsi.euclidean_op(steps+ip,&shift_iseg,igroup_iseg);
                separation=stepsi.lattice_distance();
                count++;
                if (separation==0) return(-count);
                if (separation>=min_separation) {
                    ip+=separation-min_separation+1;
                }
                else {
                    #ifdef USE_THIRD
                        sep_mod=separation%3;
                        min_separation=(2*separation+sep_mod)/3;
                        ip+=1+(separation-sep_mod)/3;
                    #else
                        sep_mod=separation%2;
                        min_separation=(separation+sep_mod)/2;
                        ip+=1+(separation-sep_mod)/2;
                    #endif
                }
            } // end loop on ip
            j-=min_separation;
            if (j<jmin-1) j=jmin-1;
        } // end decrease j
    } // end while

    return(count);

} // end test_pivot_strictly_saw()

long walk::test_pivot_bond_saw(
    long pivot_loc,long isym,excluded_class excluded)
// For the "bond SAW"
// This routine checks if the pivot produces overlapping bonds.
// It is based extremely closely on test_pivot_strictly_saw().
// We associate a bond with the index of the endpoint farthest from the
// pivot location. The condition for rejecting the pivot in other
// routine is w[jp]=w[ip]. In this routine this corresponds to
// {w[jp],w[jp+1]} = {w[ip],w[ip-1]}
// We test for equality of two bonds by checking if their midpoints are
// equal. Rather than working with the midpoints we work with twice the
// midpoints which is equal to the sum of the two endpoints.
//
// This routine only checks if the pivot produces a SAW.
// It does not carry out the pivot and it does not check if the new
// walk hits the excluded region.
{
    long i,j,ip,jp,iseg,jseg,igroup_jseg,igroup_iseg,imin,imax,jmin,jmax;
    long separation,min_separation,sep_mod;
    point origin,transi,transj,pp,stepsp,stepsi,stepsj,shift_jseg,shift_iseg,
        bondi,bondj;
    long count,changei_flag;

    // pivot is given by w[t] -> g (w[t]-w[pivot_loc])+w[pivot_loc]
    // this is equivalent to w[t] -> g w[t] + trans
    // where trans= w[pivot_loc] - g w[pivot_loc]
    // stepsp=w[pivot_loc]

    origin.zero();
    iseg=find_segment(pivot_loc,npivot,ptime);
    stepsp.euclidean_op(steps+pivot_loc,shift+iseg,igroup[iseg]);
    transi.euclidean_op(&stepsp,&origin,isym);
    transi=stepsp-transi;
    transj.euclidean_op(&stepsp,&origin,group_inverse[isym]);
    transj=stepsp-transj;

    /////////////////////////////////////////////////////////////////////////
    //
    // "Simultaneous" changing of i and j.
    // Given j<pivot_loc<i, we assume {w(jp),w(jp+1)} != {w(ip),w(ip-1)}
    // for j<jp<pivot_loc<ip<i
    // We then either decrease j or increase i.
    //
    /////////////////////////////////////////////////////////////////////////

    count=0;
    jmin=0; jmax=pivot_loc-1;
    imin=pivot_loc+1; imax=nsteps;

    j=pivot_loc-2;
    i=pivot_loc+1;
    // if pivot_loc==0, j=-2 which would allow jp=-1
    if (j<jmin-1) j=jmin-1;

    if (pivot_loc>nsteps) i=imax+1;

    while (i<=imax || j>=jmin) {
        // changei_flag=1 means we will increase i, =0 means we will increase j
        // We change the index that is closer to pivot_loc (if allowed).
        if (i-pivot_loc>pivot_loc-j) changei_flag=0;
        else changei_flag=1;
        if (i>imax) changei_flag=0;
        if (j<jmin) changei_flag=1;

        if (changei_flag) { // increase i.
	    // Need lower bound on dist from pivoted {w[i],w[i-1]} to
            // {{w[jp],w[jp+1]}: j<jp<pivot_loc}
            // This lower bound will be min_separation

            // bondi = w[i]+w[i-1]
            iseg=find_segment(i-1,npivot,ptime);
            stepsi.euclidean_op(steps+i-1,shift+iseg,igroup[iseg]);
            pp.euclidean_op(&stepsi,&transi,isym);
            bondi=pp;
            iseg=find_segment(i,npivot,ptime);
            stepsi.euclidean_op(steps+i,shift+iseg,igroup[iseg]);
            pp.euclidean_op(&stepsi,&transi,isym);
            bondi+=pp;

            min_separation=nsteps;
            jseg=npivot;   // can change to jseg=iseg; ?
            shift_jseg=shift[jseg];
            igroup_jseg=igroup[jseg];
            for (jp=jmax;jp>j;) { // note that jp is decreased

                // stepsj is w[jp+1]
                if (ptime[jseg]>jp+1) {
                    #ifdef USE_FIND_SEGMENT
                        jseg=find_segment(jp+1,npivot,ptime);
                    #else
                        while (ptime[jseg]>jp+1) jseg--;
                    #endif
                    shift_jseg=shift[jseg];
                    igroup_jseg=igroup[jseg];
                }
                stepsj.euclidean_op(steps+jp+1,&shift_jseg,igroup_jseg);
                bondj=stepsj;

                // stepsj is w[jp]
                if (ptime[jseg]>jp) {
                    #ifdef USE_FIND_SEGMENT
                        jseg=find_segment(jp,npivot,ptime);
                    #else
                        while (ptime[jseg]>jp) jseg--;
                    #endif
                    shift_jseg=shift[jseg];
                    igroup_jseg=igroup[jseg];
                }

                stepsj.euclidean_op(steps+jp,&shift_jseg,igroup_jseg);
                bondj+=stepsj;

                separation=(bondi-bondj).lattice_distance()/2;
                count++;
                if (separation==0) return(-count);
                if (separation>=min_separation) {
                    jp-=separation-min_separation+1;
                }
                else {
                    #ifdef USE_THIRD
                        sep_mod=separation%3;
                        min_separation=(2*separation+sep_mod)/3;
                        jp-=1+(separation-sep_mod)/3;
                    #else
                        sep_mod=separation%2;
                        min_separation=(separation+sep_mod)/2;
                        jp-=1+(separation-sep_mod)/2;
                    #endif
                }
            } // end loop on jp
            i+=min_separation;
            if (i>imax+1) i=imax+1;
        } // end increase i

        else {// decrease j.
            //Need lower bound on distance from {w[j],w[j+1]} to
            // pivoted {{w[ip],w[ip-1]}: pivot_loc<ip<i}
            // Equivalently we can use a lower bound on distance from inverse
            // pivoted {w[j],w[j+1]} and {{w[ip],w[ip-1]}: pivot_loc<ip<i}.
            // This lower bound will be min_separation
            // bondj = w[j]+w[j+1]
            jseg=find_segment(j,npivot,ptime);
            stepsj.euclidean_op(steps+j,shift+jseg,igroup[jseg]);
            pp.euclidean_op(&stepsj,&transj,group_inverse[isym]);
            bondj=pp;
            jseg=find_segment(j+1,npivot,ptime);
            stepsj.euclidean_op(steps+j+1,shift+jseg,igroup[jseg]);
            pp.euclidean_op(&stepsj,&transj,group_inverse[isym]);
            bondj+=pp;

            min_separation=nsteps;
            iseg=0;
            shift_iseg=shift[iseg];
            igroup_iseg=igroup[iseg];
            for (ip=imin;ip<i;) { // note that ip is increased

                // stepsi is w[ip-1]
                if (ptime[iseg+1]<=ip-1) { // check this
                    #ifdef USE_FIND_SEGMENT
                        iseg=find_segment(ip-1,npivot,ptime);
                    #else
                        while (ptime[iseg+1]<=ip-1) iseg++;
                    #endif
                    shift_iseg=shift[iseg];
                    igroup_iseg=igroup[iseg];
                }
                stepsi.euclidean_op(steps+ip-1,&shift_iseg,igroup_iseg);
                bondi=stepsi;

                // stepsi is w[ip]
                if (ptime[iseg+1]<=ip) { // check this
                    #ifdef USE_FIND_SEGMENT
                        iseg=find_segment(ip,npivot,ptime);
                    #else
                        while (ptime[iseg+1]<=ip) iseg++;
                    #endif
                    shift_iseg=shift[iseg];
                    igroup_iseg=igroup[iseg];
                }
                stepsi.euclidean_op(steps+ip,&shift_iseg,igroup_iseg);
                bondi+=stepsi;

                separation=(bondi-bondj).lattice_distance()/2;
                count++;
                if (separation==0) return(-count);
                if (separation>=min_separation) {
                    ip+=separation-min_separation+1;
                }
                else {
                    #ifdef USE_THIRD
                        sep_mod=separation%3;
                        min_separation=(2*separation+sep_mod)/3;
                        ip+=1+(separation-sep_mod)/3;
                    #else
                        sep_mod=separation%2;
                        min_separation=(separation+sep_mod)/2;
                        ip+=1+(separation-sep_mod)/2;
                    #endif
                }
            } // end loop on ip
            j-=min_separation;
            if (j<jmin-1) j=jmin-1;
        } // end decrease j
    } // end while

    return(count);

} // end test_pivot_bond_saw()

long walk::check_excluded(long pivot_loc,long isym,excluded_class excluded,
    long direction)
// checks if the proposed pivot would produce a walk that is still outside
// the excluded region.
// If the new walk hits the excluded region, routine returns -count,
// otherwise it returns count, where count is the number of
// distance computations done.
// For direction==1 the pivot is applied after pivot_loc.
// For direction==-1 the pivot is applied before pivot_loc (used by two SAW).
{
    long i,iseg;
    long separation;
    point origin,trans,pp,stepsp,stepsi;
    long count;

    // pivot is given by w[t] -> g (w[t]-w[pivot_loc])+w[pivot_loc]
    // this is equivalent to w[t] -> g w[t] + trans
    // where trans= w[pivot_loc] - g w[pivot_loc]
    // stepsp=w[pivot_loc]

    origin.zero();
    iseg=find_segment(pivot_loc,npivot,ptime);
    stepsp.euclidean_op(steps+pivot_loc,shift+iseg,igroup[iseg]);
    trans.euclidean_op(&stepsp,&origin,isym);
    trans=stepsp-trans;
    count=0;

    if (direction==1) {
        for (i=pivot_loc+1;i<=nsteps;) {
            // stepsi is w[i]
            iseg=find_segment(i,npivot,ptime);
            stepsi.euclidean_op(steps+i,shift+iseg,igroup[iseg]);
            // pp is w[i] after pivot
            pp.euclidean_op(&stepsi,&trans,isym);
            separation=pp.excluded_distance(&excluded);
            count++;
            if (separation==0) return(-count);
            i+=separation;
        } // end loop on i
        return(count);
    }

    if (direction==-1) {
        for (i=pivot_loc-1;i>=0;) {
            // stepsi is w[i]
            iseg=find_segment(i,npivot,ptime);
            stepsi.euclidean_op(steps+i,shift+iseg,igroup[iseg]);
            // pp is w[i] after pivot
            pp.euclidean_op(&stepsi,&trans,isym);
            separation=pp.excluded_distance(&excluded);
            count++;
            if (separation==0) return(-count);
            i-=separation;
        } // end loop on i
        return(count);
    }

    printf("bad direction in walk::check_excluded()\n");
    exit(1);
}

long walk::pivot_ordinary_rw(long pivot_loc,long isym,excluded_class excluded)
// pivot algorithm for the ordinary random walk. Pivot is accepted if
// the new walk does not hit the excluded region. Self-intersections do not
// matter.
{
    long iseg,count;
    point origin,transi,stepsp;

    // check if pivoted walk hits the excluded region
    count=check_excluded(pivot_loc,isym,excluded,1L);
    if (count<0) return(count);

    // If we reach this point the pivot should be accepted.
    // pivot is given by w[t] -> g (w[t]-w[pivot_loc])+w[pivot_loc]
    // this is equivalent to w[t] -> g w[t] + trans
    // where trans= w[pivot_loc] - g w[pivot_loc]
    // stepsp=w[pivot_loc]
    origin.zero();
    iseg=find_segment(pivot_loc,npivot,ptime);
    stepsp.euclidean_op(steps+pivot_loc,shift+iseg,igroup[iseg]);
    transi.euclidean_op(&stepsp,&origin,isym);
    transi=stepsp-transi;

    add_pivot(pivot_loc,isym,transi);
    return(count);

} // end pivot_ordinary_rw()


long walk::pivot_strictly_saw(long pivot_loc,long isym,excluded_class excluded)
// This is the version that changes i and j "simultaneously"
// If the pivot is accepted, this routine carries it out. Otherwise it is
// leaves the walk unchanged.
// Routine returns count if the new walk  is self-avoiding,
// returns -count if the pivot produces a self-intersection or
// intersects the excluded region,
// where count is the number of distance computations done.
// This return value is only used to study how long the routine takes.
{
    // check if pivoted walk hits the excluded region
    long count=check_excluded(pivot_loc,isym,excluded,1L);
    if (count<0) return(count);

    // check if pivoted walk is still self-avoiding
    long count_pivot=test_pivot_strictly_saw(pivot_loc,isym,excluded);
    if (count_pivot<0) return(-count+count_pivot);

    // If we reach this point the walk is self-avoiding.
    // The pivot operations were not done to the walk. So we must do them.

    // pivot is given by w[t] -> g (w[t]-w[pivot_loc])+w[pivot_loc]
    // this is equivalent to w[t] -> g w[t] + trans
    // where trans= w[pivot_loc] - g w[pivot_loc]
    // stepsp=w[pivot_loc]

    point transi,stepsp,origin;
    origin.zero();
    long iseg=find_segment(pivot_loc,npivot,ptime);
    stepsp.euclidean_op(steps+pivot_loc,shift+iseg,igroup[iseg]);
    transi.euclidean_op(&stepsp,&origin,isym);
    transi=stepsp-transi;

    add_pivot(pivot_loc,isym,transi);
    return(count+count_pivot);
} // end pivot_strictly_saw()


long walk::pivot_bond_saw(long pivot_loc,long isym,excluded_class excluded)
// Bond saw: walk cannot traverse same bond more than once.
{
    // check if pivoted walk hits the excluded region
    long count=check_excluded(pivot_loc,isym,excluded,1L);
    if (count<0) return(count);

    // check if pivoted walk is still bond self-avoiding
    long count_pivot=test_pivot_bond_saw(pivot_loc,isym,excluded);
    if (count_pivot<0) return(-count+count_pivot);

    // If we reach this point the walk is self-avoiding.
    // The pivot operations were not done to the walk. So we must do them.

    // pivot is given by w[t] -> g (w[t]-w[pivot_loc])+w[pivot_loc]
    // this is equivalent to w[t] -> g w[t] + trans
    // where trans= w[pivot_loc] - g w[pivot_loc]
    // stepsp=w[pivot_loc]

    point transi,stepsp,origin;
    origin.zero();
    long iseg=find_segment(pivot_loc,npivot,ptime);
    stepsp.euclidean_op(steps+pivot_loc,shift+iseg,igroup[iseg]);
    transi.euclidean_op(&stepsp,&origin,isym);
    transi=stepsp-transi;

    add_pivot(pivot_loc,isym,transi);
    return(count+count_pivot);
} // end pivot_bond_saw()


long walk::pivot_strictly_two_saw(
   long pivot_loc,long isym,excluded_class excluded)
// Based on pivot_strictly_saw(), but does two saw's conditioned not to
// hit each other. If pivot_loc>nsteps/2 we pivot the walk after pivot_loc.
// If pivot_loc<nsteps/2 we pivot the walk before pivot_loc.
// We do this by applying the inverse of the pivot after pivot_loc and
// then applying the pivot to all of the walk.
{
    long direction=0,count,count_pivot;

    if (pivot_loc>nsteps/2) direction=1;
    if (pivot_loc<nsteps/2) direction=-1;
    if (pivot_loc==nsteps/2) {
      if (RNG()>0.5) direction=1;
      else direction=-1;
    }

    // check if pivoted walk hits the excluded region
    count=check_excluded(pivot_loc,isym,excluded,direction);
    if (count<0) return(count);

    point trans,stepsp,origin;
    origin.zero();
    stepsp=step_rval(pivot_loc); // stepsp=w[pivot_loc]


    if (direction==1) { // pivot the walk stored in [nsteps/2,nsteps] by
        // applying g after the pivot location.
        count_pivot=test_pivot_strictly_saw(pivot_loc,isym,excluded);
        if (count_pivot<0) return(-count+count_pivot);

        // If we reach this point the walk is self-avoiding, apply pivot.
        trans.euclidean_op(&stepsp,&origin,isym);
        trans=stepsp-trans;
        add_pivot(pivot_loc,isym,trans);
        return(count+count_pivot);
    }
    else { // pivot the walk stored in [0,nsteps/2] by applying g before
        // the pivot location. We test for intersections by applying g^-1
        // after the pivot location.
        long inv_isym=group_inverse[isym];
        count_pivot=test_pivot_strictly_saw(pivot_loc,inv_isym,excluded);
        if (count_pivot<0) return(-count+count_pivot);

        // If we reach this point pivot should be accepted, apply pivot.
        // Need to apply g to walk before pivot_loc
        // We do this in two steps:
        // Step 1: apply g^-1 to walk after pivot_loc
        // Step 2: apply g to entire walk (w.r.t. w[pivot_loc])
        trans.euclidean_op(&stepsp,&origin,inv_isym);
        trans=stepsp-trans;
        add_pivot(pivot_loc,inv_isym,trans);
        trans.euclidean_op(&stepsp,&origin,isym);
        trans=stepsp-trans;
        add_pivot(0L,isym,trans);
        return(count+count_pivot);
    }
}


long walk::pivot_weakly_saw(long pivot_loc,long isym,double beta,double p,
  excluded_class excluded)
// returns 1 if the pivot is accepted
// returns 0 if the pivot is rejected and restores walk to its original state
// We accept the pivot if xrandom<exp(-beta*(new_energy-old_energy))
// where xrandom is a uniform random number in [0,1],
// and reject the pivot otherwise
// The energy of an intersection between w[i] and w[j] is beta/|i-j|^p.
// Taking p=0 gives the usual WSAW. Setting p>0 gives the "forgetful" WSAW.
{
    long i,j,ip,jp,iseg,jseg,igroup_jseg,igroup_iseg,imin,imax,jmin,jmax,
        hit_flag,separation,min_separation,sep_mod;
    point origin,trans,transi,transj,pp,stepsp,stepsi,stepsj,shift_jseg,
        shift_iseg;
    long count,changei_flag;
    double xrandom,new_energy_cutoff,old_energy,new_energy;

    // this code is very close to that in pivot_strictly_saw
    // See that code for more comments
    origin.zero();
    iseg=find_segment(pivot_loc,npivot,ptime);
    stepsp.euclidean_op(steps+pivot_loc,shift+iseg,igroup[iseg]);

    trans.euclidean_op(&stepsp,&origin,isym);
    trans=stepsp-trans;

    transi.euclidean_op(&stepsp,&origin,isym);
    transi=stepsp-transi;
    transj.euclidean_op(&stepsp,&origin,group_inverse[isym]);
    transj=stepsp-transj;

    count=0;


    // check if pivoted walk hits the excluded region
    count=check_excluded(pivot_loc,isym,excluded,1L);
    if (count<0) return(count);

    /////////////////////////////////////////////////////////////////////////
    // compute old energy
    /////////////////////////////////////////////////////////////////////////

    old_energy=0.;
    jmin=0; jmax=pivot_loc-1;
    imin=pivot_loc+1; imax=nsteps;

    j=pivot_loc-2;
    i=pivot_loc+1;
    // if pivot_loc==0, j=-2 which would allow jp=-1
    if (j<jmin-1) j=jmin-1;

    if (pivot_loc>nsteps) i=imax+1;

    while (i<=imax || j>=jmin) {
        // changei_flag=1 means we will increase i, =0 means  will increase j
        // We change the index that is closer to pivot_loc (if allowed).
        if (i-pivot_loc>pivot_loc-j) changei_flag=0;
        else changei_flag=1;
        if (i>imax) changei_flag=0;
        if (j<jmin) changei_flag=1;

        if (changei_flag) {
            // increase i. Need lower bound on dist from pivoted w[i] to
            // {w[jp]: j<jp<pivot_loc}
            // This lower bound will be min_separation
            // stepsi is w[i]
            iseg=find_segment(i,npivot,ptime);
            stepsi.euclidean_op(steps+i,shift+iseg,igroup[iseg]);
            min_separation=nsteps;
            jseg=npivot;   // can change to jseg=iseg; ?
            shift_jseg=shift[jseg]-stepsi;
            igroup_jseg=igroup[jseg];
            hit_flag=0;
            for (jp=jmax;jp>j;) { // note that j is decreased
                if (ptime[jseg]>jp) {
                    #ifdef USE_FIND_SEGMENT
                        jseg=find_segment(jp,npivot,ptime);
                    #else
                        while (ptime[jseg]>jp) jseg--;
                    #endif
                    shift_jseg=shift[jseg]-stepsi;
                    igroup_jseg=igroup[jseg];
                }
                // stepsj is w[jp]
                stepsj.euclidean_op(steps+jp,&shift_jseg,igroup_jseg);
                separation=stepsj.lattice_distance();
                count++;
                if (separation==0) {
                    hit_flag=1;
                    if (p>0.) old_energy+=pow(double(abs(jp-i)),-p);
                    else old_energy+=1.;
                    min_separation=nsteps;
                    jp--;
                }
                else { // separation>0
                    if (separation>=min_separation) {
                        jp-=separation-min_separation+1;
                    }
                    else {
                        #ifdef USE_THIRD
                            sep_mod=separation%3;
                            min_separation=(2*separation+sep_mod)/3;
                            jp-=1+(separation-sep_mod)/3;
                        #else
                            sep_mod=separation%2;
                            min_separation=(separation+sep_mod)/2;
                            jp-=1+(separation-sep_mod)/2;
                        #endif
                    }
                } // end else (separation==0)
            } // end loop on jp
            if (hit_flag) i++;
            else i+=min_separation;
            if (i>imax+1) i=imax+1;
        } // end increase i

        else {
            // decrease j. Need lower bound on distance from w[j] to
            // pivoted {w[ip]: pivot_loc<ip<i}
            // Equivalently we can use a lower bound on distance from inverse
            // pivoted w[j] and {w[ip]: pivot_loc<ip<i}
            // This lower bound will be min_separation
            // stepsj is w[j]
            jseg=find_segment(j,npivot,ptime);
            stepsj.euclidean_op(steps+j,shift+jseg,igroup[jseg]);
            min_separation=nsteps;
            iseg=0;
            shift_iseg=shift[iseg]-stepsj;
            igroup_iseg=igroup[iseg];
            hit_flag=0;
            for (ip=imin;ip<i;) { // note that i is increased
                if (ptime[iseg+1]<=ip) { // check this
                    #ifdef USE_FIND_SEGMENT
                        iseg=find_segment(ip,npivot,ptime);
                    #else
                        while (ptime[iseg+1]<=ip) iseg++;
                    #endif
                    shift_iseg=shift[iseg]-stepsj;
                    igroup_iseg=igroup[iseg];
                }
                stepsi.euclidean_op(steps+ip,&shift_iseg,igroup_iseg);
                separation=stepsi.lattice_distance();
                count++;
                if (separation==0) {
                    hit_flag=1;
                    if (p>0.) old_energy+=pow(double(abs(j-ip)),-p);
                    else old_energy+=1.;
                    min_separation=nsteps;
                    ip++;
                }
                else {
                    if (separation>=min_separation) {
                        ip+=separation-min_separation+1;
                    }
                    else {
                        #ifdef USE_THIRD
                            sep_mod=separation%3;
                            min_separation=(2*separation+sep_mod)/3;
                            ip+=1+(separation-sep_mod)/3;
                        #else
                            sep_mod=separation%2;
                            min_separation=(separation+sep_mod)/2;
                            ip+=1+(separation-sep_mod)/2;
                        #endif
                    }
                } // end else (separation==0)
            } // end loop on ip
            if (hit_flag) j--;
            else j-=min_separation;
            if (j<jmin-1) j=jmin-1;
        } // end decrease j
     } // end while

    /////////////////////////////////////////////////////////////////////////
    // compute new energy
    /////////////////////////////////////////////////////////////////////////

    // carry out the pivot and compute the new energy as we go.
    // Since the new energy can only increase, we can reject the pivot
    // once xrandom>exp(-beta*(new_energy-old_energy))  i.e.,
    // once old_energy-log(xrandom)/beta<new_energy
    xrandom=RNG();
    if (beta>1.e-10) new_energy_cutoff=old_energy-log(xrandom)/beta;
    else new_energy_cutoff=1.e10;

    new_energy=0.;
    jmin=0; jmax=pivot_loc-1;
    imin=pivot_loc+1; imax=nsteps;

    j=pivot_loc-2;
    i=pivot_loc+1;
    // if pivot_loc==0, j=-2 which would allow jp=-1
    if (j<jmin-1) j=jmin-1;

    if (pivot_loc>nsteps) i=imax+1;

    while (i<=imax || j>=jmin) {
        // changei_flag=1 means we will increase i, =0 means we will increase j
        // We change the index that is closer to pivot_loc (if allowed).
        if (i-pivot_loc>pivot_loc-j) changei_flag=0;
        else changei_flag=1;
        if (i>imax) changei_flag=0;
        if (j<jmin) changei_flag=1;

        if (changei_flag) {
            // increase i. Need lower bound on dist from pivoted w[i] to
            // {w[jp]: j<jp<pivot_loc}
            // This lower bound will be min_separation
            // stepsi is w[i]
            iseg=find_segment(i,npivot,ptime);
            stepsi.euclidean_op(steps+i,shift+iseg,igroup[iseg]);
            // pp is w[i] after pivot
            pp.euclidean_op(&stepsi,&transi,isym);
            min_separation=nsteps;
            jseg=npivot;   // can change to jseg=iseg; ?
            shift_jseg=shift[jseg]-pp;
            igroup_jseg=igroup[jseg];
            hit_flag=0;
            for (jp=jmax;jp>j;) { // note that j is decreased
                if (ptime[jseg]>jp) {
                    #ifdef USE_FIND_SEGMENT
                        jseg=find_segment(jp,npivot,ptime);
                    #else
                        while (ptime[jseg]>jp) jseg--;
                    #endif
                    shift_jseg=shift[jseg]-pp;
                    igroup_jseg=igroup[jseg];
                }
                // stepsj is w[jp]
                stepsj.euclidean_op(steps+jp,&shift_jseg,igroup_jseg);
                separation=stepsj.lattice_distance();
                count++;
                if (separation==0) {
                    hit_flag=1;
                    if (p>0.) new_energy+=pow(double(abs(jp-i)),-p);
                    else new_energy+=1.;
                    if (new_energy>new_energy_cutoff) return(-1);
                    min_separation=nsteps;
                    jp--;
                }
                else { // separation>0
                    if (separation>=min_separation) {
                        jp-=separation-min_separation+1;
                    }
                    else {
                        #ifdef USE_THIRD
                            sep_mod=separation%3;
                            min_separation=(2*separation+sep_mod)/3;
                            jp-=1+(separation-sep_mod)/3;
                        #else
                            sep_mod=separation%2;
                            min_separation=(separation+sep_mod)/2;
                            jp-=1+(separation-sep_mod)/2;
                        #endif
                    }
                } // end else (separation==0)
            } // end loop on jp
            if (hit_flag) i++;
            else i+=min_separation;
            if (i>imax+1) i=imax+1;
        } // end increase i

        else {
            // decrease j. Need lower bound on distance from w[j] to
            // pivoted {w[ip]: pivot_loc<ip<i}
            // Equivalently we can use a lower bound on distance from inverse
            // pivoted w[j] and {w[ip]: pivot_loc<ip<i}
            // This lower bound will be min_separation
            // stepsj is w[j]
            jseg=find_segment(j,npivot,ptime);
            stepsj.euclidean_op(steps+j,shift+jseg,igroup[jseg]);
            pp.euclidean_op(&stepsj,&transj,group_inverse[isym]);
            min_separation=nsteps;
            iseg=0;
            shift_iseg=shift[iseg]-pp;
            igroup_iseg=igroup[iseg];
            hit_flag=0;
            for (ip=imin;ip<i;) { // note that i is increased
                if (ptime[iseg+1]<=ip) { // check this
                    #ifdef USE_FIND_SEGMENT
                        iseg=find_segment(ip,npivot,ptime);
                    #else
                        while (ptime[iseg+1]<=ip) iseg++;
                    #endif
                    shift_iseg=shift[iseg]-pp;
                    igroup_iseg=igroup[iseg];
                }
                stepsi.euclidean_op(steps+ip,&shift_iseg,igroup_iseg);
                separation=stepsi.lattice_distance();
                count++;
                if (separation==0) {
                    hit_flag=1;
                    if (p>0.) new_energy+=pow(double(abs(j-ip)),-p);
                    else new_energy+=1.;
                    if (new_energy>new_energy_cutoff) return(-1);
                    min_separation=nsteps;
                    ip++;
                }
                else {
                    if (separation>=min_separation) {
                        ip+=separation-min_separation+1;
                    }
                    else {
                        #ifdef USE_THIRD
                            sep_mod=separation%3;
                            min_separation=(2*separation+sep_mod)/3;
                            ip+=1+(separation-sep_mod)/3;
                        #else
                            sep_mod=separation%2;
                            min_separation=(separation+sep_mod)/2;
                            ip+=1+(separation-sep_mod)/2;
                        #endif
                    }
                } // end else (separation==0)
            } // end loop on ip
            if (hit_flag) j--;
            else j-=min_separation;
            if (j<jmin-1) j=jmin-1;
        } // end decrease j
    } // end while

    // If we reach this point we accept the pivot.
    // The pivot operations were not done to the walk. So we must do them.
    add_pivot(pivot_loc,isym,trans);

    return(1);

} // end pivot_weakly_saw()

void walk::add_pivot(long pivot_loc,long isym,point trans)
{
    long iseg,ipivot;
    point pp;

    if (npivot>MAX_NPIVOT-1) {
        printf("number of implicit pivots in walk exceeds MAX_NPIVOT \n");
        exit(1);
    }

    // first, we add the pivot time to the list of pivot times.
    iseg=find_segment(pivot_loc,npivot,ptime);
    if (pivot_loc!=ptime[iseg]) {
        ptime[npivot+2]=ptime[npivot+1];
        for (ipivot=npivot;ipivot>=iseg;ipivot--) {
            ptime[ipivot+1]=ptime[ipivot];
            igroup[ipivot+1]=igroup[ipivot];
            shift[ipivot+1]=shift[ipivot];
        }
        ptime[iseg+1]=pivot_loc;
        npivot++;
        iseg++; // pivot will be applied to segments iseg to npivot
    }

    // second, we update igroup and shift
    for (ipivot=iseg;ipivot<=npivot;ipivot++) {
        pp=shift[ipivot];
        shift[ipivot].euclidean_op(&pp,&trans,isym);
        igroup[ipivot]=group_product[isym][igroup[ipivot]];
    } // end loop on ipivot

} // add_pivot()


void walk::simplify()
// carry out the pivot operations implicit in the walk, so npivot -> 0
{
    long ipivot,itime,igroup_ipivot;
    point shift_ipivot,pp;

    // even on the 0th segment there may be something to do
    for (ipivot=0;ipivot<npivot;ipivot++) {
        shift_ipivot=shift[ipivot];
        igroup_ipivot=igroup[ipivot];
        for (itime=ptime[ipivot];itime<ptime[ipivot+1];itime++) {
            pp=steps[itime];
            steps[itime].euclidean_op(&pp,&shift_ipivot,igroup_ipivot);
        }
    } // end loop on ipivot

    // npivot segment is different
    for (itime=ptime[npivot];itime<=nsteps;itime++) {
        pp=steps[itime];
        steps[itime].euclidean_op(&pp,shift+npivot,igroup[npivot]);
    }

    initialize();
} // end walk::simplify()

double walk::max_radius()
// computes max distance from walk to origin
{
    double max=0.,temp;
    for (long i=0;i<=nsteps;i++) {
        temp=steps[i].distance();
        if (temp>max) max=temp;
    }
    return(max);
}

double walk::area()
// computes area enclosed by walk that is a loop
// Green's thm says this area is (up to sign) integral of y dx and is also
// integral of x dy. We compute both to check they agree.
{
    long is;
    double x1,x2,y1,y2,area1,area2;

    // area1 is integral of x dy;  area2 is integral of y dx
    area1=0.;
    area2=0.;
    x1=steps[0].coord_x();
    y1=steps[0].coord_y();
    for (is=1;is<=nsteps;is++) {
        x2=steps[is].coord_x();
        y2=steps[is].coord_y();
        area1+=(x1+x2)*(y2-y1)/2.;
        area2+=(y1+y2)*(x2-x1)/2.;
        x1=x2;
        y1=y2;
    }

    area1=fabs(area1);
    area2=fabs(area2);
    if (fabs(area1-area2)/area1>1.e-10) {
        printf("error in area\n");
        exit(0);
    }
    return(area1);
}

double walk::diameter()
// computes diameter of walk that is a loop
{
    double delta,dist,max_dist=0.;
    long np,*p,max_np=100000,ip,jp,i,j,k;
    p = new long[max_np+1];

    // use fact=0.04;
    double fact=0.04;

    delta=fact*max_radius();
    p[0]=0;
    np=1;
    for (k=1;k<=nsteps;k++) {
      if ((steps[k]-steps[p[np-1]]).distance()>=delta) {
            if (np>=max_np-1) {
                printf("max_np too small in diameter \n");
                exit(1L);
            }
            p[np++]=k;
        }
    }
    p[np]=nsteps+1;
    double cdiam=0.;
    for (ip=0;ip<np;ip++) for (jp=ip+1;jp<np;jp++) {
        dist=(steps[p[ip]]-steps[p[jp]]).distance();
        if (dist>cdiam) cdiam=dist;
    }

    for (ip=0;ip<np;ip++) for (jp=ip+1;jp<np;jp++) {
        dist=(steps[p[ip]]-steps[p[jp]]).distance();
        if (dist>cdiam-2*delta)
        for (i=p[ip];i<p[ip+1];i++) for (j=p[jp];j<p[jp+1];j++) {
            dist=(steps[i]-steps[j]).distance();
            if (dist>max_dist) max_dist=dist;
        }
    }

    delete [] p;

    return(max_dist);
}

long bridge_inside(point p,long bridge_type,double *bridge_parms)
{
    double x,y,r;
    switch(bridge_type) {
        case 1:
            y=p.coord_y();
            if (y<bridge_parms[0]) return(1L);
            else return(0L);
        break;
        case 2:
        case -2:
            r=p.distance();
            if (r<bridge_parms[0]) return(1L);
            else return(0L);
        break;
        case 3:
            x=p.coord_x();
            y=p.coord_y();
            if (fabs(x)<bridge_parms[0] && y<bridge_parms[1]) return(1L);
            else return(0L);
        break;

        case 4:
        	x=p.coord_x();
            y=p.coord_y();
            if (y-bridge_parms[0]*(x-bridge_parms[1])>0) return(1L);
            else return(0L);

        break;

        default:
            return(0L);
        break;
    }
}


long walk::if_bridge(long bridge_type,double *bridge_parms)
{
    // the type of bridges considered is determined by a curve
    // which divides the half plane into an inside and an outside.
    // A bond is a crossing bond if it has one endpt inside and one
    // outside. We say there is a bridge if the number of
    // crossing bonds is exactly one. In this case we return i
    // where w[i],w[i+1] is the crossing bond.
    // Otherwise return -1
    // speed up: traverse walk and record I for insides sites, O for
    // outside. There is a bridge iff the resulting sequence is
    // IIIIII ... IIIIIOOOOO... OOOO
    // So for a bridge every subsequence must also be of this form.
    // So we first test some sparse subsequences to see if we can
    // quickly determine that there is no bridge.
    // determining there there is no bridge currently requires time O(N)
    // this could be sped up with work
    long i,iseg,ibridge=0;
    point p;
    long inside_flag=1;

    // preliminary test - not clear this really makes a difference in speed
    long di=long(floor(sqrt(double(nsteps))));
    for (i=0;i<=nsteps;i+=di) {
        // p is w[i]
        iseg=find_segment(i,npivot,ptime);
        p.euclidean_op(steps+i,shift+iseg,igroup[iseg]);
        if (bridge_inside(p,bridge_type,bridge_parms)) {
            if (!inside_flag) return(-1);  // OI
        }
        else {
            if (inside_flag) { // IO
                inside_flag=0;
            }
        }
    } // end loop on i

    inside_flag=1;

    for (i=0;i<=nsteps;i++) {
        // p is w[i]
        iseg=find_segment(i,npivot,ptime);
        p.euclidean_op(steps+i,shift+iseg,igroup[iseg]);
        if (bridge_inside(p,bridge_type,bridge_parms)) {
            if (!inside_flag) return(-1);  // OI
        }
        else {
            if (inside_flag) { // IO
                inside_flag=0;
                ibridge=i-1;
            }
        }
    } // end loop on i
    if (inside_flag) return(-1);
    else return(ibridge);
}


///////////////////////////////////////////////////////////////
//  checkk checkk checkk checkk checkk checkk checkk checkk  //
///////////////////////////////////////////////////////////////

